/*****************************************************************************
 * sap.c :  SAP interface module
 *****************************************************************************
 * Copyright (C) 2004-2005 the VideoLAN team
 * Copyright © 2007 Rémi Denis-Courmont
 * $Id$
 *
 * Authors: Clément Stenac <zorglub@videolan.org>
 *          Rémi Denis-Courmont
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
 *****************************************************************************/

/*****************************************************************************
 * Includes
 *****************************************************************************/
#define _GNU_SOURCE
#include <stdlib.h>                                      /* malloc(), free() */

#include <vlc/vlc.h>
#include <vlc_playlist.h>
#include <vlc_demux.h>

#include <vlc_network.h>
#include <vlc_charset.h>

#include <ctype.h>
#include <errno.h>

#ifdef HAVE_UNISTD_H
#    include <unistd.h>
#endif
#ifdef HAVE_SYS_TIME_H
#    include <sys/time.h>
#endif

#ifdef HAVE_ZLIB_H
#   include <zlib.h>
#endif

/************************************************************************
 * Macros and definitions
 ************************************************************************/

#define MAX_LINE_LENGTH 256

/* SAP is always on that port */
#define SAP_PORT 9875
/* Global-scope SAP address */
#define SAP_V4_GLOBAL_ADDRESS   "224.2.127.254"
/* Organization-local SAP address */
#define SAP_V4_ORG_ADDRESS      "239.195.255.255"
/* Local (smallest non-link-local scope) SAP address */
#define SAP_V4_LOCAL_ADDRESS    "239.255.255.255"
/* Link-local SAP address */
#define SAP_V4_LINK_ADDRESS     "224.0.0.255"
#define ADD_SESSION 1

#define SAP_V6_1 "FF0"
/* Scope is inserted between them */
#define SAP_V6_2 "::2:7FFE"
/* See RFC3513 for list of valid scopes */
/* FIXME: find a way to listen to link-local scope */
static const char ipv6_scopes[] = "1456789ABCDE";


/*****************************************************************************
 * Module descriptor
 *****************************************************************************/
#define SAP_ADDR_TEXT N_( "SAP multicast address" )
#define SAP_ADDR_LONGTEXT N_( "The SAP module normally chooses itself the " \
                              "right addresses to listen to. However, you " \
                              "can specify a specific address." )
#define SAP_IPV4_TEXT N_( "IPv4 SAP" )
#define SAP_IPV4_LONGTEXT N_( \
      "Listen to IPv4 announcements " \
      "on the standard address." )
#define SAP_IPV6_TEXT N_( "IPv6 SAP" )
#define SAP_IPV6_LONGTEXT N_( \
      "Listen to IPv6 announcements " \
      "on the standard addresses." )
#define SAP_SCOPE_TEXT N_( "IPv6 SAP scope" )
#define SAP_SCOPE_LONGTEXT N_( \
       "Scope for IPv6 announcements (default is 8)." )
#define SAP_TIMEOUT_TEXT N_( "SAP timeout (seconds)" )
#define SAP_TIMEOUT_LONGTEXT N_( \
       "Delay after which SAP items get deleted if no new announcement " \
       "is received." )
#define SAP_PARSE_TEXT N_( "Try to parse the announce" )
#define SAP_PARSE_LONGTEXT N_( \
       "This enables actual parsing of the announces by the SAP module. " \
       "Otherwise, all announcements are parsed by the \"livedotcom\" " \
       "(RTP/RTSP) module." )
#define SAP_STRICT_TEXT N_( "SAP Strict mode" )
#define SAP_STRICT_LONGTEXT N_( \
       "When this is set, the SAP parser will discard some non-compliant " \
       "announcements." )
#define SAP_CACHE_TEXT N_("Use SAP cache")
#define SAP_CACHE_LONGTEXT N_( \
       "This enables a SAP caching mechanism. " \
       "This will result in lower SAP startup time, but you could end up " \
       "with items corresponding to legacy streams." )
#define SAP_TIMESHIFT_TEXT N_("Allow timeshifting")
#define SAP_TIMESHIFT_LONGTEXT N_( "This automatically enables timeshifting " \
        "for streams discovered through SAP announcements." )

/* Callbacks */
    static int  Open ( vlc_object_t * );
    static void Close( vlc_object_t * );
    static int  OpenDemux ( vlc_object_t * );
    static void CloseDemux ( vlc_object_t * );

vlc_module_begin();
    set_shortname( _("SAP"));
    set_description( _("SAP Announcements") );
    set_category( CAT_PLAYLIST );
    set_subcategory( SUBCAT_PLAYLIST_SD );

    add_string( "sap-addr", NULL, NULL,
                SAP_ADDR_TEXT, SAP_ADDR_LONGTEXT, VLC_TRUE );
    add_bool( "sap-ipv4", 1 , NULL,
               SAP_IPV4_TEXT,SAP_IPV4_LONGTEXT, VLC_TRUE );
    add_bool( "sap-ipv6", 1 , NULL,
              SAP_IPV6_TEXT, SAP_IPV6_LONGTEXT, VLC_TRUE );
    add_integer( "sap-timeout", 1800, NULL,
                 SAP_TIMEOUT_TEXT, SAP_TIMEOUT_LONGTEXT, VLC_TRUE );
    add_bool( "sap-parse", 1 , NULL,
               SAP_PARSE_TEXT,SAP_PARSE_LONGTEXT, VLC_TRUE );
    add_bool( "sap-strict", 0 , NULL,
               SAP_STRICT_TEXT,SAP_STRICT_LONGTEXT, VLC_TRUE );
#if 0
    add_bool( "sap-cache", 0 , NULL,
               SAP_CACHE_TEXT,SAP_CACHE_LONGTEXT, VLC_TRUE );
#endif
    add_bool( "sap-timeshift", 0 , NULL,
              SAP_TIMESHIFT_TEXT,SAP_TIMESHIFT_LONGTEXT, VLC_TRUE );

    set_capability( "services_discovery", 0 );
    set_callbacks( Open, Close );

    add_submodule();
        set_description( _("SDP file parser for UDP") );
        add_shortcut( "sdp" );
        set_capability( "demux2", 51 );
        set_callbacks( OpenDemux, CloseDemux );
vlc_module_end();


/*****************************************************************************
 * Local structures
 *****************************************************************************/

typedef struct sdp_t sdp_t;
typedef struct attribute_t attribute_t;
typedef struct sap_announce_t sap_announce_t;


struct sdp_media_t
{
    struct sdp_t           *parent;
    char                   *fmt;
    struct sockaddr_storage addr;
    socklen_t               addrlen;
    unsigned                n_addr;
    int           i_attributes;
    attribute_t  **pp_attributes;
};


/* The structure that contains sdp information */
struct  sdp_t
{
    const char *psz_sdp;

    /* o field */
    char     username[64];
    uint64_t session_id;
    uint64_t session_version;
    unsigned orig_ip_version;
    char     orig_host[1024];

    /* s= field */
    char *psz_sessionname;

    /* old cruft */
    /* "computed" URI */
    char *psz_uri;
    int           i_media_type;

    /* a= global attributes */
    int           i_attributes;
    attribute_t  **pp_attributes;

    /* medias (well, we only support one atm) */
    unsigned            mediac;
    struct sdp_media_t  mediav[1];
};

struct attribute_t
{
    const char *value;
    char name[0];
};

struct sap_announce_t
{
    mtime_t i_last;

    uint16_t    i_hash;
    uint32_t    i_source[4];

    /* SAP annnounces must only contain one SDP */
    sdp_t       *p_sdp;

    int i_input_id;
    int i_item_id_cat;
    int i_item_id_one;
};

struct services_discovery_sys_t
{
    /* Socket descriptors */
    int i_fd;
    int *pi_fd;

    /* playlist node */
    playlist_item_t *p_node_cat;
    playlist_item_t *p_node_one;

    /* Table of announces */
    int i_announces;
    struct sap_announce_t **pp_announces;

    /* Modes */
    vlc_bool_t  b_strict;
    vlc_bool_t  b_parse;
    vlc_bool_t  b_timeshift;

    int i_timeout;
};

struct demux_sys_t
{
    sdp_t *p_sdp;
};

/*****************************************************************************
 * Local prototypes
 *****************************************************************************/


/* Main functions */
    static int Demux( demux_t *p_demux );
    static int Control( demux_t *, int, va_list );
    static void Run    ( services_discovery_t *p_sd );

/* Main parsing functions */
    static int ParseConnection( vlc_object_t *p_obj, sdp_t *p_sdp );
    static int ParseSAP( services_discovery_t *p_sd, const uint8_t *p_buffer, size_t i_read );
    static sdp_t *ParseSDP (vlc_object_t *p_sd, const char *psz_sdp);
    static sap_announce_t *CreateAnnounce( services_discovery_t *, uint16_t, sdp_t * );
    static int RemoveAnnounce( services_discovery_t *p_sd, sap_announce_t *p_announce );

/* Helper functions */
    static inline attribute_t *MakeAttribute (const char *str);
    static const char *GetAttribute (attribute_t **tab, unsigned n, const char *name);
    static inline void FreeAttribute (attribute_t *a);

    static vlc_bool_t IsSameSession( sdp_t *p_sdp1, sdp_t *p_sdp2 );
    static int InitSocket( services_discovery_t *p_sd, const char *psz_address, int i_port );
    static int Decompress( const unsigned char *psz_src, unsigned char **_dst, int i_len );
    static void FreeSDP( sdp_t *p_sdp );

/*****************************************************************************
 * Open: initialize and create stuff
 *****************************************************************************/
static int Open( vlc_object_t *p_this )
{
    services_discovery_t *p_sd = ( services_discovery_t* )p_this;
    services_discovery_sys_t *p_sys  = (services_discovery_sys_t *)
                                malloc( sizeof( services_discovery_sys_t ) );

    p_sys->i_timeout = var_CreateGetInteger( p_sd, "sap-timeout" );

    p_sd->pf_run = Run;
    p_sd->p_sys  = p_sys;

    p_sys->pi_fd = NULL;
    p_sys->i_fd = 0;

    p_sys->b_strict = var_CreateGetInteger( p_sd, "sap-strict");
    p_sys->b_parse = var_CreateGetInteger( p_sd, "sap-parse" );

#if 0
    if( var_CreateGetInteger( p_sd, "sap-cache" ) )
    {
        CacheLoad( p_sd );
    }
#endif

    /* Cache sap_timeshift value */
    p_sys->b_timeshift = var_CreateGetInteger( p_sd, "sap-timeshift" )
            ? VLC_TRUE : VLC_FALSE;

    /* Create our playlist node */
    pl_Yield( p_sd );

    playlist_NodesPairCreate( pl_Get( p_sd ), _("SAP sessions"),
                              &p_sys->p_node_cat, &p_sys->p_node_one,
                              VLC_TRUE );
    p_sys->p_node_cat->p_input->b_prefers_tree = VLC_TRUE;
    p_sys->i_announces = 0;
    p_sys->pp_announces = NULL;

    return VLC_SUCCESS;
}

/*****************************************************************************
 * OpenDemux: initialize and create stuff
 *****************************************************************************/
static int OpenDemux( vlc_object_t *p_this )
{
    demux_t *p_demux = (demux_t *)p_this;
    uint8_t *p_peek;
    int i_max_sdp = 1024;
    int i_sdp = 0;
    char *psz_sdp = NULL;
    sdp_t *p_sdp = NULL;

    if( !var_CreateGetInteger( p_demux, "sap-parse" ) )
    {
        /* We want livedotcom module to parse this SDP file */
        return VLC_EGENERIC;
    }

    /* Probe for SDP */
    if( p_demux->s )
    {
        if( stream_Peek( p_demux->s, &p_peek, 7 ) < 7 ) return VLC_EGENERIC;

        if( strncmp( (char*)p_peek, "v=0\r\n", 5 ) &&
            strncmp( (char*)p_peek, "v=0\n", 4 ) &&
            ( p_peek[0] < 'a' || p_peek[0] > 'z' || p_peek[1] != '=' ) )
        {
            return VLC_EGENERIC;
        }
    }

    psz_sdp = (char *)malloc( i_max_sdp );
    if( !psz_sdp ) return VLC_EGENERIC;

    /* Gather the complete sdp file */
    for( ;; )
    {
        int i_read = stream_Read( p_demux->s,
                                  &psz_sdp[i_sdp], i_max_sdp - i_sdp - 1 );

        if( i_read < 0 )
        {
            msg_Err( p_demux, "failed to read SDP" );
            goto error;
        }

        i_sdp += i_read;

        if( i_read < i_max_sdp - i_sdp - 1 )
        {
            psz_sdp[i_sdp] = '\0';
            break;
        }

        i_max_sdp += 1000;
        psz_sdp = (char *)realloc( psz_sdp, i_max_sdp );
    }

    p_sdp = ParseSDP( VLC_OBJECT(p_demux), psz_sdp );

    if( !p_sdp )
    {
        msg_Warn( p_demux, "invalid SDP");
        goto error;
    }

    if( ParseConnection( VLC_OBJECT( p_demux ), p_sdp ) )
    {
        p_sdp->psz_uri = NULL;
    }
    if( p_sdp->i_media_type != 33 && p_sdp->i_media_type != 32 &&
        p_sdp->i_media_type != 14 )
        goto error;

    if( p_sdp->psz_uri == NULL ) goto error;

    p_demux->p_sys = (demux_sys_t *)malloc( sizeof(demux_sys_t) );
    p_demux->p_sys->p_sdp = p_sdp;
    p_demux->pf_control = Control;
    p_demux->pf_demux = Demux;

    FREENULL( psz_sdp );
    return VLC_SUCCESS;

error:
    FREENULL( psz_sdp );
    if( p_sdp ) FreeSDP( p_sdp ); p_sdp = NULL;
    stream_Seek( p_demux->s, 0 );
    return VLC_EGENERIC;
}

/*****************************************************************************
 * Close:
 *****************************************************************************/
static void Close( vlc_object_t *p_this )
{
    services_discovery_t *p_sd = ( services_discovery_t* )p_this;
    services_discovery_sys_t    *p_sys  = p_sd->p_sys;

    int i;

    for( i = p_sys->i_fd-1 ; i >= 0 ; i-- )
    {
        net_Close( p_sys->pi_fd[i] );
    }
    FREENULL( p_sys->pi_fd );

#if 0
    if( config_GetInt( p_sd, "sap-cache" ) )
    {
        CacheSave( p_sd );
    }
#endif

    for( i = p_sys->i_announces  - 1;  i>= 0; i-- )
    {
        RemoveAnnounce( p_sd, p_sys->pp_announces[i] );
    }
    FREENULL( p_sys->pp_announces );

    playlist_NodeDelete( pl_Get(p_sd), p_sys->p_node_cat, VLC_TRUE,
                         VLC_TRUE );
    playlist_NodeDelete( pl_Get(p_sd), p_sys->p_node_one, VLC_TRUE,
                         VLC_TRUE );
    pl_Release( p_sd );
    free( p_sys );
}

/*****************************************************************************
 * CloseDemux: Close the demuxer
 *****************************************************************************/
static void CloseDemux( vlc_object_t *p_this )
{
    demux_t *p_demux = (demux_t *)p_this;
    if( p_demux->p_sys )
    {
        if( p_demux->p_sys->p_sdp ) { FreeSDP( p_demux->p_sys->p_sdp ); p_demux->p_sys->p_sdp = NULL; }
        free( p_demux->p_sys );
    }
}

/*****************************************************************************
 * Run: main SAP thread
 *****************************************************************************
 * Listens to SAP packets, and sends them to packet_handle
 *****************************************************************************/
#define MAX_SAP_BUFFER 5000

static void Run( services_discovery_t *p_sd )
{
    char *psz_addr;
    int i;

    /* Braindead Winsock DNS resolver will get stuck over 2 seconds per failed
     * DNS queries, even if the DNS server returns an error with milliseconds.
     * You don't want to know why the bug (as of XP SP2) wasn't fixed since
     * Winsock 1.1 from Windows 95, if not Windows 3.1.
     * Anyway, to avoid a 30 seconds delay for failed IPv6 socket creation,
     * we have to open sockets in Run() rather than Open(). */
    if( var_CreateGetInteger( p_sd, "sap-ipv4" ) )
    {
        InitSocket( p_sd, SAP_V4_GLOBAL_ADDRESS, SAP_PORT );
        InitSocket( p_sd, SAP_V4_ORG_ADDRESS, SAP_PORT );
        InitSocket( p_sd, SAP_V4_LOCAL_ADDRESS, SAP_PORT );
        InitSocket( p_sd, SAP_V4_LINK_ADDRESS, SAP_PORT );
    }
    if( var_CreateGetInteger( p_sd, "sap-ipv6" ) )
    {
        char psz_address[] = SAP_V6_1"0"SAP_V6_2;
        const char *c_scope;

        for( c_scope = ipv6_scopes; *c_scope; c_scope++ )
        {
            psz_address[sizeof(SAP_V6_1) - 1] = *c_scope;
            InitSocket( p_sd, psz_address, SAP_PORT );
        }
    }

    psz_addr = var_CreateGetString( p_sd, "sap-addr" );
    if( psz_addr && *psz_addr )
    {
        InitSocket( p_sd, psz_addr, SAP_PORT );
        free( psz_addr );
    }

    if( p_sd->p_sys->i_fd == 0 )
    {
        msg_Err( p_sd, "unable to listen on any address" );
        return;
    }

    /* read SAP packets */
    while( !p_sd->b_die )
    {
        int i_read;
        uint8_t p_buffer[MAX_SAP_BUFFER+1];

        i_read = net_Select( p_sd, p_sd->p_sys->pi_fd,
                             p_sd->p_sys->i_fd, p_buffer,
                             MAX_SAP_BUFFER, 500000 );

        /* Check for items that need deletion */
        for( i = 0; i < p_sd->p_sys->i_announces; i++ )
        {
            mtime_t i_timeout = ( mtime_t ) 1000000 * p_sd->p_sys->i_timeout;

            if( mdate() - p_sd->p_sys->pp_announces[i]->i_last > i_timeout )
            {
                RemoveAnnounce( p_sd, p_sd->p_sys->pp_announces[i] );
            }
        }

        /* Minimum length is > 6 */
        if( i_read <= 6 )
        {
            if( i_read < 0 )
            {
                msg_Warn( p_sd, "socket read error" );
            }
            continue;
        }

        p_buffer[i_read] = '\0';

        /* Parse the packet */
        ParseSAP( p_sd, p_buffer, i_read );
    }
}

/**********************************************************************
 * Demux: reads and demuxes data packets
 * Return -1 if error, 0 if EOF, 1 else
 **********************************************************************/
static int Demux( demux_t *p_demux )
{
    sdp_t *p_sdp = p_demux->p_sys->p_sdp;
    input_thread_t *p_input;
    input_item_t *p_parent_input;

    playlist_t *p_playlist = pl_Yield( p_demux );
    p_input = (input_thread_t *)vlc_object_find( p_demux, VLC_OBJECT_INPUT,
                                                 FIND_PARENT );
    assert( p_input );
    if( !p_input )
    {
        msg_Err( p_demux, "parent input could not be found" );
        return VLC_EGENERIC;
    }

    p_parent_input = input_GetItem(p_input);

    vlc_mutex_lock( &p_parent_input->lock );
    FREENULL( p_parent_input->psz_uri );
    p_parent_input->psz_uri = strdup( p_sdp->psz_uri );
    FREENULL( p_parent_input->psz_name );
    p_parent_input->psz_name = strdup( p_sdp->psz_sessionname );
    p_parent_input->i_type = ITEM_TYPE_NET;

    if( p_playlist->status.p_item &&
             p_playlist->status.p_item->p_input == p_parent_input )
    {
        playlist_Control( p_playlist, PLAYLIST_VIEWPLAY, VLC_TRUE,
                          p_playlist->status.p_node, p_playlist->status.p_item );
    }

    vlc_mutex_unlock( &p_parent_input->lock );
    vlc_object_release( p_input );
    vlc_object_release( p_playlist );

    return VLC_SUCCESS;
}

static int Control( demux_t *p_demux, int i_query, va_list args )
{
    return VLC_EGENERIC;
}

/**************************************************************
 * Local functions
 **************************************************************/

/* i_read is at least > 6 */
static int ParseSAP( services_discovery_t *p_sd, const uint8_t *buf,
                     size_t len )
{
    int i;
    const char          *psz_sdp;
    const uint8_t *end = buf + len;
    sdp_t               *p_sdp;

    assert (buf[len] == '\0');

    if (len < 4)
        return VLC_EGENERIC;

    uint8_t flags = buf[0];

    /* First, check the sap announce is correct */
    if ((flags >> 5) != 1)
        return VLC_EGENERIC;

    vlc_bool_t b_ipv6 = (flags & 0x10) != 0;
    vlc_bool_t b_need_delete = (flags & 0x04) != 0;

    if (flags & 0x02)
    {
        msg_Dbg( p_sd, "encrypted packet, unsupported" );
        return VLC_EGENERIC;
    }

    vlc_bool_t b_compressed = (flags & 0x01) != 0;

    uint16_t i_hash = U16_AT (buf + 2);

    if( p_sd->p_sys->b_strict && i_hash == 0 )
    {
        msg_Dbg( p_sd, "strict mode, discarding announce with null id hash");
        return VLC_EGENERIC;
    }

    // Skips source address and auth data
    buf += 4 + (b_ipv6 ? 16 : 4) + buf[1];
    if (buf > end)
        return VLC_EGENERIC;

    uint8_t *decomp = NULL;
    if( b_compressed )
    {
        int newsize = Decompress (buf, &decomp, end - buf);
        if (newsize < 0)
        {
            msg_Dbg( p_sd, "decompression of SAP packet failed" );
            return VLC_EGENERIC;
        }

        decomp = realloc (decomp, newsize + 1);
        decomp[newsize] = '\0';
        psz_sdp = (const char *)decomp;
        len = newsize;
    }
    else
    {
        psz_sdp = (const char *)buf;
        len = end - buf;
    }

    /* len is a strlen here here. both buf and decomp are len+1 where the 1 should be a \0 */
    assert( psz_sdp[len] == '\0');

    /* Skip payload type */
    /* SAPv1 has implicit "application/sdp" payload type: first line is v=0 */
    if (strncmp (psz_sdp, "v=0", 3))
    {
        size_t clen = strlen (psz_sdp) + 1;

        if (strcmp (psz_sdp, "application/sdp"))
        {
            msg_Dbg (p_sd, "unsupported content type: %s", psz_sdp);
            return VLC_EGENERIC;
        }

        // skips content type
        if (len <= clen)
            return VLC_EGENERIC;

        len -= clen;
        psz_sdp += clen;
    }

    /* Parse SDP info */
    p_sdp = ParseSDP( VLC_OBJECT(p_sd), psz_sdp );

    if( p_sdp == NULL )
        return VLC_EGENERIC;

    p_sdp->psz_sdp = psz_sdp;

    /* Decide whether we should add a playlist item for this SDP */
    /* Parse connection information (c= & m= ) */
    if( ParseConnection( VLC_OBJECT(p_sd), p_sdp ) )
        p_sdp->psz_uri = NULL;

    /* Multi-media or no-parse -> pass to LIVE.COM */
    if( ( p_sdp->i_media_type != 14
       && p_sdp->i_media_type != 32
       && p_sdp->i_media_type != 33)
     || p_sd->p_sys->b_parse == VLC_FALSE )
    {
        free( p_sdp->psz_uri );
        if (asprintf( &p_sdp->psz_uri, "sdp://%s", p_sdp->psz_sdp ) == -1)
            p_sdp->psz_uri = NULL;
    }

    if( p_sdp->psz_uri == NULL ) return VLC_EGENERIC;

    for( i = 0 ; i< p_sd->p_sys->i_announces ; i++ )
    {
        /* FIXME: slow */
        /* FIXME: we create a new announce each time the sdp changes */
        if( IsSameSession( p_sd->p_sys->pp_announces[i]->p_sdp, p_sdp ) )
        {
            if( b_need_delete )
            {
                RemoveAnnounce( p_sd, p_sd->p_sys->pp_announces[i]);
            }
            else
            {
                p_sd->p_sys->pp_announces[i]->i_last = mdate();
            }
            FreeSDP( p_sdp ); p_sdp = NULL;
            return VLC_SUCCESS;
        }
    }

    CreateAnnounce( p_sd, i_hash, p_sdp );

    FREENULL (decomp);
    return VLC_SUCCESS;
}

sap_announce_t *CreateAnnounce( services_discovery_t *p_sd, uint16_t i_hash,
                                sdp_t *p_sdp )
{
    input_item_t *p_input;
    playlist_item_t     *p_item, *p_child;
    const char *psz_value;
    sap_announce_t *p_sap = (sap_announce_t *)malloc(
                                        sizeof(sap_announce_t ) );
    services_discovery_sys_t *p_sys;
    if( p_sap == NULL )
        return NULL;

    p_sys = p_sd->p_sys;

    p_sap->i_last = mdate();
    p_sap->i_hash = i_hash;
    p_sap->p_sdp = p_sdp;

    /* Create the actual playlist item here */
    p_input = input_ItemNewWithType( VLC_OBJECT(p_sd),
                                     p_sap->p_sdp->psz_uri,
                                     p_sdp->psz_sessionname,
                                     0, NULL, -1, ITEM_TYPE_NET );
    p_sap->i_input_id = p_input->i_id;
    if( !p_input )
    {
        free( p_sap );
        return NULL;
    }

    if( p_sys->b_timeshift )
        input_ItemAddOption( p_input, ":access-filter=timeshift" );

    psz_value = GetAttribute( p_sap->p_sdp->pp_attributes, p_sap->p_sdp->i_attributes, "tool" );
    if( psz_value != NULL )
    {
        input_ItemAddInfo( p_input, _("Session"),_("Tool"), psz_value );
    }
    if( strcmp( p_sdp->username, "-" ) )
    {
        input_ItemAddInfo( p_input, _("Session"),
                                _("User"), p_sdp->username );
    }

    /* Handle group */
    psz_value = GetAttribute( p_sap->p_sdp->pp_attributes, p_sap->p_sdp->i_attributes, "x-plgroup" );
    if( psz_value == NULL )
        psz_value = GetAttribute( p_sap->p_sdp->pp_attributes, p_sap->p_sdp->i_attributes, "plgroup" );

    if( psz_value != NULL )
    {
        p_child = playlist_ChildSearchName( p_sys->p_node_cat, psz_value );

        if( p_child == NULL )
        {
            p_child = playlist_NodeCreate( pl_Get( p_sd ), psz_value,
                                           p_sys->p_node_cat, 0 );
            p_child->i_flags &= ~PLAYLIST_SKIP_FLAG;
        }
    }
    else
    {
        p_child = p_sys->p_node_cat;
    }

    p_item = playlist_NodeAddInput( pl_Get( p_sd ), p_input, p_child,
                                    PLAYLIST_APPEND, PLAYLIST_END );
    p_item->i_flags &= ~PLAYLIST_SKIP_FLAG;
    p_item->i_flags &= ~PLAYLIST_SAVE_FLAG;
    p_sap->i_item_id_cat = p_item->i_id;

    p_item = playlist_NodeAddInput( pl_Get( p_sd ), p_input,
                        p_sys->p_node_one, PLAYLIST_APPEND, PLAYLIST_END );
    p_item->i_flags &= ~PLAYLIST_SKIP_FLAG;
    p_item->i_flags &= ~PLAYLIST_SAVE_FLAG;
    p_sap->i_item_id_one = p_item->i_id;

    TAB_APPEND( p_sys->i_announces, p_sys->pp_announces, p_sap );

    return p_sap;
}


static const char *FindAttribute (const sdp_t *sdp, unsigned media,
                                  const char *name)
{
    /* Look for media attribute, and fallback to session */
    return GetAttribute (sdp->mediav[media].pp_attributes,
                         sdp->mediav[media].i_attributes, name)
        ?: GetAttribute (sdp->pp_attributes, sdp->i_attributes, name);
}


/* Fill p_sdp->psz_uri */
static int ParseConnection( vlc_object_t *p_obj, sdp_t *p_sdp )
{
    if (p_sdp->mediac != 1)
        return VLC_EGENERIC;

    char psz_uri[1026];
    const char *host;
    int port;

    psz_uri[0] = '[';
    if (vlc_getnameinfo ((struct sockaddr *)&(p_sdp->mediav[0].addr),
                         p_sdp->mediav[0].addrlen, psz_uri + 1,
                         sizeof (psz_uri) - 2, &port, NI_NUMERICHOST))
        return VLC_EGENERIC;

    if (strchr (psz_uri + 1, ':'))
    {
        host = psz_uri;
        psz_uri[strlen (psz_uri)] = ']';
    }
    else
        host = psz_uri + 1;

    /* Parse m= field */
    char *sdp_proto = strdup (p_sdp->mediav[0].fmt);
    if (sdp_proto == NULL)
        return VLC_ENOMEM;

    char *subtype = strchr (sdp_proto, ' ');
    if (sdp_proto == NULL)
    {
        msg_Dbg (p_obj, "missing SDP media subtype: %s", sdp_proto);
        p_sdp->i_media_type = 0;
    }
    else
    {
        *subtype++ = '\0';
        p_sdp->i_media_type = atoi (subtype);
    }
    if (p_sdp->i_media_type == 0)
         p_sdp->i_media_type = 33;

    /* RTP protocol, nul, VLC shortcut, nul, flags byte as follow:
     * 0x1: Connection-Oriented media. */
    static const char proto_match[] =
        "udp\0"             "udp\0\0"
        "RTP/AVP\0"         "rtp\0\0"
        "UDPLite/RTP/AVP\0" "udplite\0\0"
        "DCCP/RTP/AVP\0"    "dccp\0\1"
        "TCP/RTP/AVP\0"     "rtptcp\0\1"
        "\0";

    const char *vlc_proto = NULL;
    for (const char *proto = proto_match; *proto;)
    {
        if (strcasecmp (proto, sdp_proto))
        {
            vlc_proto = proto + strlen (proto) + 1;

            if (vlc_proto[strlen (vlc_proto) + 1] & 1)
            {
                /* COMEDIA check */
                const char *setup = FindAttribute (p_sdp, 0, "setup");
                if (setup == NULL)
                    setup = "active"; /* default value */

                if (strcmp (setup, "actpass") && strcmp (setup, "passive"))
                {
                    msg_Dbg (p_obj, "unsupported COMEDIA mode: %s", setup);
                    free (sdp_proto);
                    return VLC_EGENERIC;
                }
            }
            break;
        }
        proto += strlen (proto) + 1;
        proto += strlen (proto) + 2;
    }

    free (sdp_proto);
    if (vlc_proto == NULL)
    {
        msg_Dbg (p_obj, "unknown SDP media protocol: %s",
                 p_sdp->mediav[0].fmt);
        return VLC_EGENERIC;
    }

    /* handle SSM case */
    const char *sfilter = FindAttribute (p_sdp, 0, "source-filter");
    char psz_source[258] = "";
    if (sfilter != NULL)
    {
        char psz_source_ip[256];

        if (sscanf (sfilter, " incl IN IP%*c %*s %255s ", psz_source_ip) == 1)
        {
            if (strchr (psz_source_ip, ':') != NULL)
                sprintf (psz_source, "[%s]", psz_source_ip);
            else
                strcpy (psz_source, psz_source_ip);
        }
    }

    asprintf( &p_sdp->psz_uri, "%s://%s@%s:%i", vlc_proto, psz_source,
              host, port );

    return VLC_SUCCESS;
}


static int ParseSDPConnection (const char *str, struct sockaddr_storage *addr,
                               socklen_t *addrlen, unsigned *number)
{
    char host[60];
    unsigned fam, n1, n2;

    int res = sscanf (str, "IN IP%u %59[^/]/%u/%u", &fam, host, &n1, &n2);
    if (res < 2)
        return -1;

    switch (fam)
    {
#ifdef AF_INET6
        case 6:
            addr->ss_family = AF_INET6;
# ifdef HAVE_SA_LEN
            addr->ss_len =
# endif
           *addrlen = sizeof (struct sockaddr_in6);

            if (inet_pton (AF_INET6, host,
                           &((struct sockaddr_in6 *)addr)->sin6_addr) <= 0)
                return -1;

            *number = (res >= 3) ? n1 : 1;
            break;
#endif

        case 4:
            addr->ss_family = AF_INET;
# ifdef HAVE_SA_LEN
            addr->ss_len =
# endif
           *addrlen = sizeof (struct sockaddr_in);

            if (inet_pton (AF_INET, host,
                           &((struct sockaddr_in *)addr)->sin_addr) <= 0)
                return -1;

            *number = (res >= 4) ? n2 : 1;
            break;

        default:
            return -1;
    }
    return 0;
}


/***********************************************************************
 * ParseSDP : SDP parsing
 * *********************************************************************
 * Validate SDP and parse all fields
 ***********************************************************************/
static sdp_t *ParseSDP (vlc_object_t *p_obj, const char *psz_sdp)
{
    if( psz_sdp == NULL )
        return NULL;

    sdp_t *p_sdp = calloc (1, sizeof (*p_sdp));
    if (p_sdp == NULL)
        return NULL;

    char expect = 'V';
    struct sockaddr_storage glob_addr;
    memset (&glob_addr, 0, sizeof (glob_addr));
    socklen_t glob_len = 0;
    unsigned glob_count = 1;

    /* TODO: use iconv and charset attribute instead of EnsureUTF8 */
    while (*psz_sdp)
    {
        /* Extract one line */
        char *eol = strchr (psz_sdp, '\n');
        size_t linelen = eol ? (size_t)(eol - psz_sdp) : strlen (psz_sdp);
        char line[linelen + 1];
        memcpy (line, psz_sdp, linelen);
        line[linelen] = '\0';

        psz_sdp += linelen + 1;

        /* Remove carriage return if present */
        eol = strchr (line, '\r');
        if (eol != NULL)
        {
            linelen = eol - line;
            line[linelen] = '\0';
        }

        /* Validate line */
        char cat = line[0], *data = line + 2;
        if (!cat || (strchr ("vosiuepcbtrzkam", cat) == NULL))
        {
            /* MUST ignore SDP with unknown line type */
            msg_Dbg (p_obj, "unknown SDP line type: 0x%02x", (int)cat);
            goto error;
        }
        if (line[1] != '=')
        {
            msg_Dbg (p_obj, "invalid SDP line: %s", line);
            goto error;
        }

        assert (linelen >= 2);

        /* SDP parsing state machine
         * We INTERNALLY use uppercase for session, lowercase for media
         */
        switch (expect)
        {
            /* Session description */
            case 'V':
                expect = 'O';
                if (cat != 'v')
                {
                    msg_Dbg (p_obj, "missing SDP version");
                    goto error;
                }
                if (strcmp (data, "0"))
                {
                    msg_Dbg (p_obj, "unknown SDP version: %s", data);
                    goto error;
                }
                break;

            case 'O':
            {
                expect = 'S';
                if (cat != 'o')
                {
                    msg_Dbg (p_obj, "missing SDP originator");
                    goto error;
                }

                if ((sscanf (data, "%63s "I64Fu" "I64Fu" IN IP%u %1023s",
                             p_sdp->username, &p_sdp->session_id,
                             &p_sdp->session_version, &p_sdp->orig_ip_version,
                             p_sdp->orig_host) != 5)
                 || ((p_sdp->orig_ip_version != 4)
                  && (p_sdp->orig_ip_version != 6)))
                {
                    msg_Dbg (p_obj, "SDP origin not supported: %s\n", data);
                    /* Or maybe out-of-range, but this looks suspicious */
                    return NULL;
                }
                EnsureUTF8 (p_sdp->orig_host);
                break;
            }

            case 'S':
            {
                expect = 'I';
                if ((cat != 's') || !*data)
                {
                    /* MUST be present AND non-empty */
                    msg_Dbg (p_obj, "missing SDP session name");
                    goto error;
                }
                assert (p_sdp->psz_sessionname == NULL); // no memleak here
                p_sdp->psz_sessionname = strdup (data);
                EnsureUTF8 (p_sdp->psz_sessionname);
                if (p_sdp->psz_sessionname == NULL)
                    goto error;
                break;
            }

            case 'I':
                expect = 'U';
                if (cat == 'i')
                    break;
            case 'U':
                expect = 'E';
                if (cat == 'u')
                    break;
            case 'E':
                expect = 'E';
                if (cat == 'e')
                    break;
            case 'P':
                expect = 'P';
                if (cat == 'p')
                    break;
            case 'C':
                expect = 'B';
                if (cat == 'c')
                {
                    if (ParseSDPConnection (data, &glob_addr, &glob_len,
                                            &glob_count))
                    {
                        msg_Dbg (p_obj, "SDP connection infos not supported: "
                                 "%s", data);
                        goto error;
                    }
                    break;
                }
            case 'B':
                assert (expect == 'B');
                if (cat == 'b')
                    break;
            case 'T':
                expect = 'R';
                if (cat != 't')
                {
                    msg_Dbg (p_obj, "missing SDP time description");
                    goto error;
                }
                break;

            case 'R':
                if ((cat == 't') || (cat == 'r'))
                    break;

            case 'Z':
                expect = 'K';
                if (cat == 'z')
                    break;
            case 'K':
                expect = 'A';
                if (cat == 'k')
                    break;
            case 'A':
                //expect = 'A';
                if (cat == 'a')
                {
                    attribute_t *p_attr = MakeAttribute (data);
                    TAB_APPEND( p_sdp->i_attributes, p_sdp->pp_attributes, p_attr );
                    break;
                }

            /* Media description */
            case 'm':
            {
                expect = 'i';
                if (cat != 'm')
                {
                    msg_Dbg (p_obj, "missing SDP media description");
                    goto error;
                }
                struct sdp_media_t *m = p_sdp->mediav + p_sdp->mediac;

                memcpy (&m->addr, &glob_addr, m->addrlen = glob_len);
                m->n_addr = glob_count;

                /* TODO: remember media type (if we need multiple medias) */
                data = strchr (data, ' ');
                if (data == NULL)
                {
                    msg_Dbg (p_obj, "missing SDP media port");
                    goto error;
                }
                int port = atoi (++data);
                if (port <= 0 || port >= 65536)
                {
                    msg_Dbg (p_obj, "invalid transport port %d", port);
                    goto error;
                }
                net_SetPort ((struct sockaddr *)&m->addr, htons (port));

                data = strchr (data, ' ');
                if (data == NULL)
                {
                    msg_Dbg (p_obj, "missing SDP media format");
                    goto error;
                }
                m->fmt = strdup (++data);
                if (m->fmt == NULL)
                    goto error;

                p_sdp->mediac++;
                break;
            }
            case 'i':
                expect = 'c';
                if (cat == 'i')
                    break;
            case 'c':
                expect = 'b';
                if (cat == 'c')
                {
                    struct sdp_media_t *m = p_sdp->mediav + p_sdp->mediac;
                    if (ParseSDPConnection (data, &m->addr, &m->addrlen,
                                            &m->n_addr))
                    {
                        msg_Dbg (p_obj, "SDP connection infos not supported: "
                                 "%s", data);
                        goto error;
                    }
                    break;
                }
            case 'b':
                expect = 'b';
                if (cat == 'b')
                    break;
            case 'k':
                expect = 'a';
                if (cat == 'k')
                    break;
            case 'a':
                assert (expect == 'a');
                if (cat == 'a')
                {
                    attribute_t *p_attr = MakeAttribute (data);
                    if (p_attr == NULL)
                        goto error;

                    TAB_APPEND (p_sdp->mediav[p_sdp->mediac - 1].i_attributes,
                                p_sdp->mediav[p_sdp->mediac - 1].pp_attributes, p_attr);
                    break;
                }

                if (cat == 'm')
                {
                    /* TODO */
                    msg_Dbg (p_obj, "multi-media SDP not implemented -> live555");
                    goto error;
                }

                if (cat != 'm')
                {
                    msg_Dbg (p_obj, "unexpected SDP line: 0x%02x", (int)cat);
                    goto error;
                }
                break;

            default:
                msg_Err (p_obj, "*** BUG in SDP parser! ***");
                goto error;
        }
    }

    return p_sdp;

error:
    FreeSDP (p_sdp);
    return NULL;
}

static int InitSocket( services_discovery_t *p_sd, const char *psz_address,
                       int i_port )
{
    int i_fd = net_ListenUDP1 ((vlc_object_t *)p_sd, psz_address, i_port);
    if (i_fd == -1)
        return VLC_EGENERIC;

    net_StopSend( i_fd );
    INSERT_ELEM (p_sd->p_sys->pi_fd, p_sd->p_sys->i_fd,
                 p_sd->p_sys->i_fd, i_fd);
    return VLC_SUCCESS;
}

static int Decompress( const unsigned char *psz_src, unsigned char **_dst, int i_len )
{
#ifdef HAVE_ZLIB_H
    int i_result, i_dstsize, n = 0;
    unsigned char *psz_dst = NULL;
    z_stream d_stream;

    memset (&d_stream, 0, sizeof (d_stream));

    i_result = inflateInit(&d_stream);
    if( i_result != Z_OK )
        return( -1 );

    d_stream.next_in = (Bytef *)psz_src;
    d_stream.avail_in = i_len;

    do
    {
        n++;
        psz_dst = (unsigned char *)realloc( psz_dst, n * 1000 );
        d_stream.next_out = (Bytef *)&psz_dst[(n - 1) * 1000];
        d_stream.avail_out = 1000;

        i_result = inflate(&d_stream, Z_NO_FLUSH);
        if( ( i_result != Z_OK ) && ( i_result != Z_STREAM_END ) )
        {
            inflateEnd( &d_stream );
            return( -1 );
        }
    }
    while( ( d_stream.avail_out == 0 ) && ( d_stream.avail_in != 0 ) &&
           ( i_result != Z_STREAM_END ) );

    i_dstsize = d_stream.total_out;
    inflateEnd( &d_stream );

    *_dst = (unsigned char *)realloc( psz_dst, i_dstsize );

    return i_dstsize;
#else
    (void)psz_src;
    (void)_dst;
    (void)i_len;
    return -1;
#endif
}


static void FreeSDP( sdp_t *p_sdp )
{
    free( p_sdp->psz_sessionname );
    free( p_sdp->psz_uri );

    for (unsigned j = 0; j < p_sdp->mediac; j++)
    {
        free (p_sdp->mediav[j].fmt);
        for (int i = 0; i < p_sdp->mediav[j].i_attributes; i++)
            FreeAttribute (p_sdp->mediav[j].pp_attributes[i]);
    }

    for (int i = 0; i < p_sdp->i_attributes; i++)
        FreeAttribute (p_sdp->pp_attributes[i]);

    free (p_sdp->pp_attributes);
    free (p_sdp);
}

static int RemoveAnnounce( services_discovery_t *p_sd,
                           sap_announce_t *p_announce )
{
    int i;

    if( p_announce->p_sdp )
    {
        FreeSDP( p_announce->p_sdp );
        p_announce->p_sdp = NULL;
    }

    if( p_announce->i_input_id > -1 )
        playlist_DeleteFromInput( pl_Get(p_sd), p_announce->i_input_id, VLC_FALSE );

    for( i = 0; i< p_sd->p_sys->i_announces; i++)
    {
        if( p_sd->p_sys->pp_announces[i] == p_announce )
        {
            REMOVE_ELEM( p_sd->p_sys->pp_announces, p_sd->p_sys->i_announces,
                         i);
            break;
        }
    }

    free( p_announce );

    return VLC_SUCCESS;
}

static vlc_bool_t IsSameSession( sdp_t *p_sdp1, sdp_t *p_sdp2 )
{
    /* A session is identified by
     * - username,
     * - session_id,
     * - network type (which is always IN),
     * - address type (currently, this means IP version),
     * - and hostname.
     */
    if (strcmp (p_sdp1->username, p_sdp2->username)
     || (p_sdp1->session_id != p_sdp2->session_id)
     || (p_sdp1->orig_ip_version != p_sdp2->orig_ip_version)
     || strcmp (p_sdp1->orig_host, p_sdp2->orig_host))
        return VLC_FALSE;

    return VLC_TRUE;
}


static inline attribute_t *MakeAttribute (const char *str)
{
    attribute_t *a = malloc (sizeof (*a) + strlen (str) + 1);
    if (a == NULL)
        return NULL;

    strcpy (a->name, str);
    EnsureUTF8 (a->name);
    char *value = strchr (a->name, ':');
    if (value != NULL)
    {
        *value++ = '\0';
        a->value = value;
    }
    else
        a->value = "";
    return a;
}


static const char *GetAttribute (attribute_t **tab, unsigned n,
                                 const char *name)
{
    for (unsigned i = 0; i < n; i++)
        if (strcasecmp (tab[i]->name, name) == 0)
            return tab[i]->value;
    return NULL;
}


static inline void FreeAttribute (attribute_t *a)
{
    free (a);
}
